package main

import (
	"fmt"
	"io"
	"mime"
	"net"
	"net/url"
	"os"
	"path"
)

type Status int

const (
	Input                     = 10
	Success                   = 20
	Redirect                  = 30
	RedirectPermanent         = 31
	TemporaryFailure          = 40
	PermanentFailure          = 50
	NotFound                  = 51
	ProxyRequestRefused       = 53
	BadRequest                = 59
	ClientCertificateRequired = 60
)

const ResponseEnd = "\r\n"
const MaxURLSize = 1024
const IndexFile = "index.gmi"
const GeminiMIME = "text/gemini"

func handleRequest(c net.Conn, di int, parsedURL *url.URL) {
	if parsedURL.Path == "" {
		parsedURL.Path = "/"
		redirectPermanent(c, parsedURL.String())
		return
	} else if parsedURL.Path != path.Clean(parsedURL.Path) {
		sendError(c, BadRequest, "Path error")
		return
	}

	if parsedURL.Path == "/" || parsedURL.Path == "." {
		serve(c, di, IndexFile)
	} else {
		serve(c, di, parsedURL.Path)
	}
}

func serve(c net.Conn, di int, filepath string) {
	fullPath := path.Join(config.Domain[di].Root, filepath)

	pathInfo, err := os.Stat(fullPath)
	if err != nil {
		sendError(c, NotFound, "Not found")
		return
	}

	if pathInfo.IsDir() {
		subDirIndex := path.Join(fullPath, IndexFile)
		if _, err := os.Stat(subDirIndex); os.IsNotExist(err) {
			sendError(c, NotFound, "Not found")
			return
		}

		fullPath = subDirIndex
	}

	mimeType := getMimeType(fullPath)

	if mimeType == "" {
		sendError(c, NotFound, "")
		return
	}

	file, err := os.Open(fullPath)
	if err != nil {
		sendError(c, NotFound, "")
		return
	}
	defer file.Close()

	success(c, file, mimeType)
}

func redirectPermanent(c net.Conn, url string) {
	c.Write(getResponseBytes(RedirectPermanent, url))
}

func sendError(c net.Conn, code int, message string) {
	c.Write(getResponseBytes(code, message))
}

func success(c net.Conn, file *os.File, mimeType string) {
	c.Write(getResponseBytes(Success, mimeType))
	io.Copy(c, file)
}

func getResponseBytes(code int, message string) []byte {
	if message == "" {
		return []byte(fmt.Sprintf("%d%s", code, ResponseEnd))
	}
	return []byte(fmt.Sprintf("%d %s%s", code, message, ResponseEnd))
}

func getMimeType(fullPath string) string {
	ext := path.Ext(fullPath)
	if ext == ".gmi" || ext == ".gemini" {
		return GeminiMIME
	}
	return mime.TypeByExtension(ext)
}
